define(['jquery', 'util', 'rangy-core', 'rangy-cssclassapplier'], function ($, util, rangy) {
	var orig_text = '', clean_text = '', exceptions = '.,?!:();-%+', menu_columns = 1, maxcolumns = 10, appliers = {};
	check_regex = /<sel(?:[^"'>]|"([^"]*)"|'[^']*')*>(\S*?)<\/sel>/gmi,
			clean_regex = /<\/?sel(?:[^"'>]|"[^"]*"|'[^']*')*>/gmi,
			xhtml_clean = /xmlns="(.*?)"/gmi;
	var create_styles = function (accents) {
		var l = accents.length, style_text = '',
				style = document.createElement('style');
		style.type = 'text/css';
		style.id = 'tmp_style';
		accents.each(function () {
			var t = $(this), cn = t.attr('id') || t.attr('data_id');
			style_text += '.accent #main_text .sel_' + cn + ' { color: ' + t.attr('color').replace('0x', '#') + '; } ';
			appliers['sel_' + cn] = rangy.createCssClassApplier('sel_' + cn, {normalize: true, elementTagName: 'sel'});
		});
		if (style.styleSheet)
			style.styleSheet.cssText = style_text;
		else
			style.appendChild(document.createTextNode(style_text));

		document.getElementsByTagName('head')[0].appendChild(style);
	},
			parse_data = function (data, state) {
				exceptions = data.attr('exceptions') || '.,?!:();-%+';
				menu_columns = data.attr('menu_columns') || 1;
				menu_columns = menu_columns > maxcolumns ? maxcolumns : menu_columns;

				var colors = data.find('accents sel'), l = colors.length, col_width = 99 / menu_columns,
						res = '<div class="col" style="width_:' + col_width + '%">',
						per_col = Math.ceil((l + 1) / menu_columns);
				colors.each(function (ind) {
					var t = $(this), cn = t.attr('id') || t.attr('data_id');
					res += '<div id="sel_' + cn + '" style="background-color:' + t.attr('color').replace('0x', '#') + '"><span><i></i>' + util.get_html($(this)) + '</span></div>';
					//if (ind%per_col == 0) {
					res += '</div><div class="col" style="width_:' + col_width + '%">';
					//}
				});
				res += '<div id="sel_no"><span><i></i>сброс выделения</span></div></div>';
				orig_text = util.get_html(data.find('text'));
				clean_text = orig_text.replace(clean_regex, '');
				var font = data.attr('text_size');
				return '<div class="accent"><div id="main_text" ' + (font ? 'style="font-size:' + font + 'px"' : '') + '>' + (state ? state : clean_text) + '</div><div class="colors">' + res + '</div></div>';
			},
			attach_events = function (elem) {
				$(elem).find('.accent .colors .col span').on('mousedown touchstart', function () {
					var el = $(this).parent(), type = el.attr('id') || el.attr('data_id');
					$.each(appliers, function (sel, applier) {
						if (applier)
							applier.undoToSelection();
					});
					if (type != 'sel_no') {
						if (appliers[type])
							appliers[type].applyToSelection();
					}
					rangy.getSelection().removeAllRanges();
					if ($('#main_text sel').length > 0) {
						$('#check_btn').removeClass('inactive');
					} else {
						$('#check_btn').addClass('inactive');
					}
					return false;
				});
			},
			get_text = function (el) {
				return $('<div>' + el.text() + '</div>').text();
			}

	return {
		/* elem - element, where the data is rendered
		 data - <component> element to render
		 state - current state (if the question is already answered) */
		init_question: function (elem, data, state, r, cb) {
			maxcolumns = Math.floor(($(elem).width() - 50) / 100);
			$(elem).html(parse_data(data, state));
			var cel = $('.accent .colors')[0], add = cel.lastChild.offsetTop - cel.firstChild.offsetTop;
			$('#main_text').css('max-height', parseInt($(elem).css('max-height')) - 81 - add + 'px');
			//$(elem).html(q_text);
			rangy.init();
			create_styles(data.find('accents sel'));
			attach_events(elem);
			if (typeof cb == 'function') {
				cb();
			}
		},
		/* cd - callback, function(result,score,state),
		 result - failed, passed, partial (-1, 1, 0), score - 0-100, state - state to save and resore when returned to this question */
		process_answer: function (cb) {
			if ($('#main_text sel').length == 0) {
				return false;
			}
			var num_result = 0;
			$.each(appliers, function (sel, applier) {
				appliers[sel] = null;
			});

			var ex_array = exceptions.split(''), i = ex_array.length, text_correct = orig_text, state = text_check = $('#main_text').html(),
					chunks_correct = [], chunks_check = [], chunks_correct_col = [], chunks_check_col = [], match,
					num_correct = 0, num_incorect = 0, num_selected = 0, num_orig = 0;
			while (i--) {
				text_correct = text_correct.split(ex_array[i]).join('');
				text_check = text_check.split(ex_array[i]).join('');
			}
			text_correct = text_correct.replace(/[\s\n\r\t]/gm, '');
			text_check = text_check.replace(/[\s\n\r\t]/gm, '').replace(xhtml_clean, '');

			while (match = check_regex.exec(text_correct)) {
				if (match[2].length > 0) {
					var last = chunks_correct.length;
					if (last > 0 && match[1] == chunks_correct[last - 1].class_name && check_regex.lastIndex - match[0].length == text_correct.length - chunks_correct[last - 1].after_length) {
						chunks_correct[last - 1].after = text_correct.substr(check_regex.lastIndex).replace(clean_regex, '');
						chunks_correct[last - 1].after_length = text_correct.substr(check_regex.lastIndex).length;
						chunks_correct[last - 1].texts.push(match[2]);
						chunks_correct[last - 1].inds.push(num_orig);
					} else {
						chunks_correct.push({
							beafore_raw: text_correct.substr(0, check_regex.lastIndex - match[0].length),
							before: text_correct.substr(0, check_regex.lastIndex - match[0].length).replace(clean_regex, ''),
							after_length: text_correct.substr(check_regex.lastIndex).length,
							after: text_correct.substr(check_regex.lastIndex).replace(clean_regex, ''),
							class_name: match[1],
							texts: [match[2]],
							inds: [num_orig]
						});
					}
				}
				num_orig++;
			}
			while (match = check_regex.exec(text_check)) {
				if (match[2].length > 0) {
					var last = chunks_check.length;
					if (last > 0 && match[1].replace('sel_', '') == chunks_check[last - 1].class_name && check_regex.lastIndex - match[0].length == text_check.length - chunks_check[last - 1].after_length) {
						chunks_check[last - 1].after = text_check.substr(check_regex.lastIndex).replace(clean_regex, '');
						chunks_check[last - 1].after_length = text_check.substr(check_regex.lastIndex).length;
						chunks_check[last - 1].texts.push(match[2]);
						chunks_check[last - 1].inds.push(num_selected);
					} else {
						chunks_check.push({
							before: text_check.substr(0, check_regex.lastIndex - match[0].length).replace(clean_regex, ''),
							after_length: text_check.substr(check_regex.lastIndex).length,
							after: text_check.substr(check_regex.lastIndex).replace(clean_regex, ''),
							class_name: match[1].replace('sel_', ''),
							texts: [match[2]],
							inds: [num_selected]
						});
					}
				}
				num_selected++;
			}
			//var num_selected = chunks_check.length, num_orig = chunks_correct.length;
			for (var i = 0; i < chunks_check.length; i++) {
				var is_correct = 0;
				for (var j = 0; j < chunks_correct.length; j++) {
					if (chunks_check[i].before == chunks_correct[j].before &&
							chunks_check[i].after == chunks_correct[j].after &&
							chunks_check[i].class_name == chunks_correct[j].class_name) {
						is_correct = 1;
						num_correct += chunks_correct[j].texts.length;
						num_selected += chunks_correct[j].texts.length - chunks_check[i].texts.length;
						break;
					} else if (chunks_correct[j].texts.length > 0 && chunks_check[i].class_name == chunks_correct[j].class_name) {
						var tmp_before = chunks_correct[j].before, ll = chunks_correct[j].texts.length;
						for (var ii = 0; ii < ll; ii++) {
							var tmp_after = chunks_correct[j].after;
							for (var jj = ll - 1; jj >= ii; jj--) {
								if (chunks_check[i].before == tmp_before && chunks_check[i].after == tmp_after) {
									is_correct = 1;
									num_correct += jj + 1 - ii;
									num_selected += jj + 1 - ii - chunks_check[i].texts.length;
									break;
								}
								tmp_after = chunks_correct[j].texts[jj] + tmp_after;
							}
							tmp_before = tmp_before + chunks_correct[j].texts[ii];
						}
					}
				}
				if (!is_correct) {
					var jj = chunks_check[i].inds.length;
					while (jj--) {
						if (chunks_check[i].texts[jj]) {
							$($('#main_text sel')[chunks_check[i].inds[jj]]).addClass('error');
							num_incorect++;
						}
					}
				}
			}
			num_result = num_correct * 2 - num_selected;
			num_result = num_result < 0 ? 0 : num_result * 100 / num_orig;

			cb(num_result < 100 ? (num_incorect == 0 ? 0 : -1) : 1, num_result, state);
			return true;
		},
		show_correct: function () {
			$('#main_text').html(orig_text.replace(/<sel id="/g, '<sel class="sel_').replace(/<sel data_id="/g, '<sel class="sel_'));
		}
	};
});
