define(['jquery', 'util', 'kinetic'], function ($, util, kinetic) {

	Kinetic.Stage.prototype._setPointerPosition = function (evt) {
		var contentPosition = this._getContentPosition(),
				offsetX = evt.offsetX,
				clientX = evt.clientX,
				x = null,
				y = null,
				touch;
		evt = evt ? evt : window.event;

		// touch events
		if (evt.touches !== undefined) {
			// currently, only handle one finger
			if (evt.touches.length > 0) {

				touch = evt.touches[0];

				// get the information for finger #1
				x = (touch.clientX - contentPosition.left) / window['gScale'];
				y = (touch.clientY - contentPosition.top) / window['gScale'];

			}
		}
		// mouse events
		else {
			// if offsetX is defined, assume that offsetY is defined as well
			if (offsetX !== undefined) {
				x = offsetX;
				y = evt.offsetY;
			}
			// we unfortunately have to use UA detection here because accessing
			// the layerX or layerY properties in newer versions of Chrome
			// throws a JS warning.  layerX and layerY are required for FF
			// when the container is transformed via CSS.
			else if (Kinetic.UA.browser === 'mozilla') {

				x = evt.layerX || (evt.clientX - contentPosition.left);
				y = evt.layerY || (evt.clientY - contentPosition.top);
			}
			// if clientX is defined, assume that clientY is defined as well
			else if (clientX !== undefined && contentPosition) {
				x = (clientX - contentPosition.left);
				y = evt.clientY - contentPosition.top;
			}
		}
		if (x !== null && y !== null) {
			this.pointerPos = {
				x: x,
				y: y
			};
		}
	}

	var line, binding, bind_array = [], moving = false, snapped = false, startcol = -1, bindings_check, canvas_active,
			canvas = [],
			main_color = '#287dad', main_color_dot = '#0ff1ff', correct_color = '#0fff37', incorrect_color = '#ff0e0e', corrected_color = '#aaaaaa',
			lineWidth = 3, inshift = 23;
	hitFunc = function (context) {
		var hw = 10, shift = 10, p = this.points(),
				l = Math.sqrt((p[2] - p[0]) * (p[2] - p[0]) + (p[3] - p[1]) * (p[3] - p[1])),
				c = (p[2] - p[0]) / l, s = (p[3] - p[1]) / l;
		context.beginPath();
		context.moveTo(p[0] + shift * c + hw * s, p[1] + shift * s - hw * c);
		context.lineTo(p[2] - shift * c + hw * s, p[3] - shift * s - hw * c);
		context.lineTo(p[2] - shift * c - hw * s, p[3] - shift * s + hw * c);
		context.lineTo(p[0] + shift * c - hw * s, p[1] + shift * s + hw * c);
		context.closePath();
		context.fillStrokeShape(this);
	},
			draw_circle = function (x, y, layer, rectpos) {
				if (rectpos == 'left') {
					var r = new Kinetic.Rect({
						x: x - 40,
						y: y - 23,
						width: 40,
						height: 46,
						fill: '#ffffff',
						stroke: 'transparent'
					});
				} else {
					var r = new Kinetic.Rect({
						x: x,
						y: y - 23,
						width: 40,
						height: 46,
						fill: '#ffffff',
						stroke: 'transparent'
					});
				}
				var circ1 = new Kinetic.Circle({
					x: x,
					y: y,
					radius: 23,
					stroke: 'transparent',
					strokeWidth: 0,
					fill: '#ffffff'
				}), circ2 = new Kinetic.Circle({
					x: x,
					y: y,
					radius: 14,
					stroke: 'transparent',
					strokeWidth: 0,
					fillLinearGradientStartPoint: {x: 0, y: 14},
					fillLinearGradientEndPoint: {x: 0, y: -14},
					fillLinearGradientColorStops: [0, '#1c5a85', 1, '#3093c7']
				}), circ3 = new Kinetic.Circle({
					x: x,
					y: y,
					radius: 7,
					stroke: 'transparent',
					strokeWidth: 0,
					fill: '#ffffff',
					hitFunc: hit_circle
				});
				//layer.add(r);
				//layer.add(circ1);
				layer.add(circ2);
				layer.add(circ3);
				return circ3;
			},
			hit_circle = function (context) {
				context.beginPath();
				context.arc(0, 0, 20, 0, 2 * Math.PI, false);
				context.closePath();
				context.fillStrokeShape(this);
			},
			md_circle = function (evt, cobj, stage, layer) {
				if (canvas_active) {
					//evt.preventDefault();
					evt.target.setFill(main_color_dot);

					startcol = cobj.col;
					if (line) {
						line.destroy();
					}
					line = new Kinetic.Line({
						points: [evt.target.x(), evt.target.y(), evt.target.x(), evt.target.y()],
						stroke: main_color,
						strokeWidth: lineWidth,
						hitFunc: hitFunc
					});

					layer.add(line);
					var mousePos = stage.getPointerPosition();
					moving = true;
					evt.target.moveToTop();
					layer.drawScene();
					binding = {line: line, from: cobj.id, from_circ: evt.target};
				}

			},
			mo_circle = function (evt, cobj, stage, layer, circles) {
				//evt.preventDefault();
				if (snapped && moving && circles[cobj.id].bonds <= 0) {
					evt.target.setFill('#ffffff');
					layer.drawScene();
				}
				snapped = false;
			},
			mov_circle = function (evt, cobj, stage, layer) {
				if (canvas_active && moving && cobj.col != startcol && line) {
					//evt.preventDefault();
					snapped = true;
					evt.target.setFill(main_color_dot);
					evt.target.moveToTop();
					var p = line.points();
					p[2] = evt.target.x();
					p[3] = evt.target.y();
					line.points(p);
					layer.drawScene();
				}
			},
			mu_circle = function (evt, cobj, stage, layer, circles) {
				if (canvas_active && moving && cobj.col != startcol && line) {
					//evt.preventDefault();
					evt.target.setFill(main_color_dot);
					evt.target.moveToTop();
					moving = false;
					startcol = -1;
					var p = line.points();
					p[2] = evt.target.x();
					p[3] = evt.target.y();
					line.points(p);
					layer.draw();
					binding.to = cobj.id;
					binding.to_circ = evt.target;
					circles[binding.to].bonds++;
					circles[binding.from].bonds++;
					bind_array.push(binding);
					var bl = bind_array.length - 1;
					line.on('click tap', function (evt) {
						if (canvas_active) {
							evt.target.destroy();
							if (--circles[bind_array[bl].from].bonds <= 0)
								circles[bind_array[bl].from].el.setFill('#ffffff');
							if (--circles[bind_array[bl].to].bonds <= 0)
								circles[bind_array[bl].to].el.setFill('#ffffff');
							layer.draw();
							bind_array[bl] = null;
							var i = bind_array.length, binds_exist = false;
							while (i--) {
								binds_exist = binds_exist || !!bind_array[i];
							}
							if (!binds_exist) {
								$('#check_btn').addClass('inactive');
							}
						}
					});
					line = null;
					$('#check_btn').removeClass('inactive');
				}
			},
			init_canvas = function (elem_id, circles1, circles2, state) {
				var elem = $('#' + elem_id), colwidth = elem.width(), colheight = elem.height(), circles = {}, extent = 20;
				var layer = new Kinetic.Layer();
				var stage = new Kinetic.Stage({
					container: elem_id,
					width: colwidth + extent * 2,
					height: colheight
				});
				var background = new Kinetic.Rect({
					x: 0,
					y: 0,
					width: stage.getWidth(),
					height: stage.getHeight(),
					fill: "transparent"
				});

				layer.add(background);
				stage.add(layer);

				for (var i = 0; i < circles1.length; i++) {
					var circ = draw_circle(inshift + extent, circles1[i].pos - 1, layer, 'left');
					(function (cobj) {
						circ.on("mousedown touchstart", function (evt) {
							md_circle(evt, cobj, stage, layer);
							if(!$('#check_btn').hasClass('inactive')){
								reset_colors();
							}
						});
						circ.on("mouseover touchenter", function (evt) {
							mov_circle(evt, cobj, stage, layer)
						});
						circ.on("mouseout touchleave", function (evt) {
							mo_circle(evt, cobj, stage, layer, circles)
						});
						circ.on("mouseup touchend", function (evt) {
							mu_circle(evt, cobj, stage, layer, circles)
						});
					})(circles1[i]);
					circles[circles1[i].id] = {el: circ, id: circles1[i].id, bonds: 0};
				}
				for (var i = 0; i < circles2.length; i++) {
					var circ = draw_circle(stage.getWidth() - inshift - extent, circles2[i].pos - 1, layer, 'right');
					(function (cobj) {
						circ.on("mousedown touchstart", function (evt) {
							md_circle(evt, cobj, stage, layer);
							if(!$('#check_btn').hasClass('inactive')){
								reset_colors();
							}
						});
						circ.on("mouseover touchenter", function (evt) {
							mov_circle(evt, cobj, stage, layer)
						});
						circ.on("mouseout touchleave", function (evt) {
							mo_circle(evt, cobj, stage, layer, circles)
						});
						circ.on("mouseup touchend", function (evt) {
							mu_circle(evt, cobj, stage, layer, circles)
						});
					})(circles2[i]);
					circles[circles2[i].id] = {el: circ, id: circles2[i].id, bonds: 0};
				}
				if (state) {
					//bind_array = [];
					var sl = state.length;
					for (i = 0; i < sl; i++) {
						var from = circles[state[i].from];
						var to = circles[state[i].to];
						if (from && to) {
							var sline = new Kinetic.Line({
								points: [from.el.x(), from.el.y(), to.el.x(), to.el.y()],
								stroke: main_color,
								strokeWidth: lineWidth,
								hitFunc: hitFunc
							});
							layer.add(sline);
							bind_array.push({from: state[i].from, to: state[i].to, from_circ: from.el, to_circ: to.el, line: sline});
						}
					}
				}
				layer.draw();

				stage.on("mousemove touchmove", function (evt) {
					if (canvas_active && moving && !snapped && line) {
						//evt.preventDefault();
						var mousePos = stage.getPointerPosition();
						var x = mousePos.x;
						var y = mousePos.y;
						var points = line.points();
						points[2] = x;
						points[3] = y;
						line.points(points);
						moving = true;
						layer.drawScene();
					}
				});

				stage.on("mouseout touchleave", function (evt) {
					if (canvas_active && moving && line && evt.evt.type == 'mouseout') {
						line.destroy();
						if (circles[binding.from].bonds <= 0)
							circles[binding.from].el.setFill('#ffffff');
						layer.draw();
						line = null;
					}
				});
				stage.on("mouseup touchend", function (evt) {
					if (canvas_active && moving && line) {
						//evt.preventDefault();
						line.destroy();
						if (circles[binding.from].bonds <= 0)
							circles[binding.from].el.setFill('#ffffff');
						layer.draw();
						line = null;
					}
				});
				return {el: layer, circles: circles};
			},
			reset_colors = function () {
				var i = bind_array.length;
				bindings_check.attr('done', 0);
				while (i--) {
					var el = bind_array[i], is_correct = false;
					if (el) {
						bindings_check.each(function () {
							var id1 = $(this).attr('id1'), id2 = $(this).attr('id2');
							if ((el.from == id1 && el.to == id2) || (el.from == id2 && el.to == id1)) {
								is_correct = true;
								$(this).attr('done', 1);
							}
						});
						el.line.stroke(main_color);
						el.from_circ.setFill(main_color_dot);
						el.from_circ.moveToTop();
						el.to_circ.setFill(main_color_dot);
						el.to_circ.moveToTop();
						el.line.getLayer().drawScene();
					}
				}

			}
	return {
		/* elem - element, where the data is rendered
		 data - <component> element to render
		 state - current state (if the question is already answered) */
		init_question: function (elem, data, state, renderer, cb) {
			var cols = data.find('columns').find('column'), ncols = cols.length,
					colwidth = ($(elem).width() - 1 - (ncols * 2 - 2) * 46) / (ncols * 2 - 1), colheight = parseInt($(elem).css('max-height')) - 1,
					sepwidth = (parseInt(data.attr('ColumnSpace')) || colwidth) + 92,
					img_max_width = parseInt(data.attr('pic_w')) || 10000,
					image_array = [], cols_array = [];
			colwidth = ($(elem).width() - 1 - (ncols - 1) * sepwidth) / ncols;
			$(elem).css('overflow-y', 'visible');
			$(elem).css('overflow-x', 'visible');
			bind_array = [];
			cols.each(function (ic) {
				var col = {title: $(this).find('title').text(), elems: []},
				statements = $(this).find('statement');
				if ($(this).attr('mix') != 'no') {
					if (!state) {
						statements = util.array_shuffle(statements);
						$(this).data('stored_data', statements);
					} else {
						statements = $(this).data('stored_data');
					}
				}
				statements.each(function (ie) {
					var t = $(this);
					if (t.attr('picture')) {
						image_array.push({col: ic, el: ie, sid: t.attr('id') || t.attr('data_id'), src: t.attr('picture')});
						col.elems.push({id: t.attr('id') || t.attr('data_id'), image: image_array[image_array.length - 1], text: t.text()});
					} else {
						col.elems.push({id: t.attr('id') || t.attr('data_id'), text: t.text()})
					}
				});
				cols_array.push(col);
			});
			bindings_check = data.find('bond');
			bind_array = [];
			canvas = [];
			util.load_images(image_array, function () {
				for (var i = 0; i < ncols; i++) {
					var col_el = $('<div>').addClass('col').css('width', colwidth + 'px'), col = (cols_array[i].title == '' ? '<h3></h3>' : '<h3 class="border">' + cols_array[i].title + '</h3>'),
							cl = cols_array[i].elems.length, cumheight = 0;
					for (var j = 0; j < cl; j++) {
						var el = cols_array[i].elems[j];
						if (el.image) {
							col += '<div class="img_wrap"><div style="' + (img_max_width ? 'max-width:' + img_max_width + 'px;' : '') + 'width:' + colwidth + 'px;" class="img">' + renderer.render_image(el.image.src, true, el.text, 'auto', '100%') + '</div></div>';
							cumheight += el.image.height / el.image.width * Math.min(colwidth, img_max_width);
						} else {
							col += '<div style="width:' + colwidth + 'px;" class="txt"><div class="text">' + el.text + '</div></div>';
						}
					}
					col_el.html(col);
					cols_array[i].el = col_el;
					cols_array[i].cumheight = cumheight;
				}
				var all_cols = $('<div>').addClass('connection');
				for (var i = 0; i < ncols; i++) {
					all_cols.append(cols_array[i].el);
					if (i < ncols - 1) {
						all_cols.append('<div class="col"><div id="canvas_' + i + '" style="width:' + sepwidth + 'px; height:' + colheight + 'px;"></div></div>');
					}
				}

				$(elem).html(all_cols);
				for (var i = 0; i < ncols; i++) {
					if (cols_array[i].el.height() > colheight) {
						var newheight = colheight - cols_array[i].el.height() + cols_array[i].cumheight, coef = newheight / cols_array[i].cumheight;
						var imgs = $(cols_array[i].el).find('.img');
						imgs.css('width', (Math.min(colwidth, img_max_width) * coef) + 'px');
						if (i == 0) {
							//imgs.css('margin-left',(colwidth*(1-coef))+'px');
						} else if (i == ncols - 1) {
							//imgs.css('margin-right',(colwidth*(1-coef))+'px');
						}
						if (newheight < 0)
							$('<div>').addClass('error_message').appendTo(elem).html('ВНИМАНИЕ! Данные для столбца ' + (i + 1) + ' не помещаются на экран');
					} else {
						var pad = (colheight - cols_array[i].el.height());
						if (pad > (cols_array[i].el.find('.text').length - 1) * 33) {
							cols_array[i].el.find('.txt').css({'margin-bottom': '46px'});
							//pad = (colheight - cols_array[i].el.height())/2;
						}
						//cols_array[i].el.find('h3').css('margin-bottom',pad+'px');
					}
				}
				var circles = [];
				for (var i = 0; i < ncols; i++) {
					var circ = [];
					$(cols_array[i].el).children('div').each(function (ind) {
						circ.push({col: i, id: cols_array[i].elems[ind].id, pos: this.offsetTop - this.parentNode.offsetTop + $(this).height() / 2});
					});
					circles.push(circ);
					if (i > 0) {
						canvas.push(init_canvas('canvas_' + (i - 1), circles[i - 1], circles[i], state));
					}
				}
				if (state) {
					canvas_active = false;
				} else {
					canvas_active = true;
				}
				if (typeof cb == 'function') {
					cb();
				}
			});
		},
		/* cd - callback, function(result,score,state),
		 result - failed, passed, partial (-1, 1, 0), score - 0-100, state - state to save and resore when returned to this question */
		process_answer: function (cb, forceNotAllChoosedCheck, forceNotDoneCheck) {
			console.log('forceNotAllChoosedCheck ' + forceNotAllChoosedCheck);
			if (typeof forceNotAllChoosedCheck == 'undefined') {
				forceNotAllChoosedCheck = true;
			}
			if (typeof forceNotDoneCheck == 'undefined') {
				forceNotDoneCheck = false;
			}
			console.log(forceNotDoneCheck);
			var num_result = 0, num_incorrect = 0, state = [], i = bind_array.length, binds_exist = false;
			while (i--) {
				binds_exist = binds_exist || !!bind_array[i];
			}
			if (!binds_exist && !forceNotDoneCheck) {
				return false;
			}
			if (!binds_exist && forceNotDoneCheck) {
				cb(-1, 0, state);
				return true;
			}
			i = bind_array.length - 1;
			canvas_active = false;
			while (i--) {
				var j = i + 1, el1 = bind_array[i + 1];
				if (el1)
					while (j--) {
						var el2 = bind_array[j];
						if (el2 && ((el1.from == el2.from && el1.to == el2.to) || (el1.from == el2.to && el1.to == el2.from))) {
							bind_array[j] = null;
						}
					}

			}
			var i = bind_array.length;
			bindings_check.attr('done', 0);
			while (i--) {
				var el = bind_array[i], is_correct = false;
				if (el) {
					state.push({from: el.from, to: el.to});
					bindings_check.each(function () {
						var id1 = $(this).attr('id1'), id2 = $(this).attr('id2');
						if ((el.from == id1 && el.to == id2) || (el.from == id2 && el.to == id1)) {
							is_correct = true;
							$(this).attr('done', 1);
						}
					});
					if (is_correct) {
						num_result++;
						el.line.stroke(correct_color);
						el.from_circ.setFill(correct_color);
						el.from_circ.moveToTop();
						el.to_circ.setFill(correct_color);
						el.to_circ.moveToTop();
						el.state = 1;
					} else {
						num_result--;
						num_incorrect++;
						el.line.stroke(incorrect_color);
						el.from_circ.setFill(incorrect_color);
						el.from_circ.moveToTop();
						el.to_circ.setFill(incorrect_color);
						el.to_circ.moveToTop();
						el.state = 0;
					}
					el.line.getLayer().drawScene();
				}
			}
			num_result = num_result < 0 ? 0 : num_result * 100 / bindings_check.length;
			cb(num_result < 100 ? (num_incorrect == 0 && !forceNotAllChoosedCheck ? 0 : -1) : 1, num_result, state);
			return true;
		},
		show_correct: function () {
			var cl = canvas.length;
			var i = bind_array.length;
			while (i--) {
				if (bind_array[i].state == 0) {
					bind_array[i].line.destroy();
					bind_array[i].from_circ.setFill('#ffffff');
					bind_array[i].to_circ.setFill('#ffffff');
				}
			}
			i = bind_array.length;
			while (i--) {
				if (bind_array[i].state == 1) {
					bind_array[i].from_circ.setFill(correct_color);
					bind_array[i].to_circ.setFill(correct_color);
				}
			}
			bindings_check.each(function () {
				//if ($(this).attr('done') != 1) {
				var id1 = $(this).attr('id1'), id2 = $(this).attr('id2'), curlayer = false;
				for (var i = 0; i < cl; i++) {
					if (typeof canvas[i].circles[id1] != 'undefined' && typeof canvas[i].circles[id2] != 'undefined') {
						curlayer = canvas[i];
					}
				}
				var line = new Kinetic.Line({
					points: [curlayer.circles[id1].el.x(), curlayer.circles[id1].el.y(), curlayer.circles[id2].el.x(), curlayer.circles[id2].el.y()],
					stroke: correct_color,
					strokeWidth: lineWidth
				});

				curlayer.el.add(line);
				curlayer.circles[id1].el.setFill(correct_color);
				curlayer.circles[id1].el.moveToTop();
				curlayer.circles[id2].el.setFill(correct_color);
				curlayer.circles[id2].el.moveToTop();
				curlayer.el.drawScene();
				//}
			});
		},
		enable_answers: function () {
			canvas_active = true;
		},
		disable_answers: function () {
			canvas_active = false;
		},
//		reset_highlight: function () {
//			var l = correct.length;
//			for (var i = 0; i < l; i++) {
//				cnt.find('select[data-id=' + i + ']').removeClass('incorrect correct').selectmenu('instance').button.removeClass('incorrect correct');
//			}
//		},
		get_correct_answer_text: function () {
			var curr_task = app_data.tasks[app_data.cur_task];

			function cloneCanvas(oldCanvas) {

				//create a new canvas
				var newCanvas = document.createElement('canvas');
				var context = newCanvas.getContext('2d');

				//set dimensions
				newCanvas.width = oldCanvas.width;
				newCanvas.height = oldCanvas.height;

				//apply the old canvas to the new one
				context.drawImage(oldCanvas, 0, 0);

				//return the new canvas
				return newCanvas;
			}

			function cloneQuestion($question) {
				var width = $question.width();
				var $content = $question.clone(),
						canvases = [];
				$question.find('[id^=canvas_]').each(function () {
					canvases.push(cloneCanvas($(this).find('canvas')[0]));
				});

				$content.find('[id^=canvas_]').each(function (index) {
					$(this).find('canvas').replaceWith($(canvases[index]));
				});
				$content.width(width);
				return $content;

			}

			if (typeof curr_task.correctQuestionWithCanvas == 'undefined') {
				var $oldQuestion = cloneQuestion($('#question_cnt'));
				this.show_correct();
				var $correctQ = cloneQuestion($('#question_cnt'));
				$('#question_cnt').replaceWith($oldQuestion);
				curr_task.correctQuestionWithCanvas = $correctQ;
				return $correctQ;
			} else {
				return curr_task.correctQuestionWithCanvas
			}
		}
	};
});