define(['jquery', 'util'], function ($, util) {
	var inputs = [],
			restrictions = [
				/[0-9]/g,
				/[А-Яа-яЁё]/g,
				/[A-Za-z]/g
			],
			layouts = ['my-num', 'my-russian', 'my-english'],
			current_layout = 0,
			parse_string = function (txt, state) {
				var res = '', arr = txt.replace(/&#/g, '[[!--ampdash--]]').split('#'), len = arr.length - 1;
				inputs = [];
				var max_max_length = 0;
				for (var i = 0; i < len; i += 2) {
					inputs[i / 2] = {variants: arr[i + 1].split('~'), state: 0};
					var maxlength = util.longest_string(inputs[i / 2].variants).length + 2;
					if(max_max_length < maxlength){
						max_max_length = maxlength;
					}
				}
				for (var i = 0; i < len; i += 2) {
					res += arr[i].replace(/\[\[\!--ampdash--\]\]/g, '&#');
					inputs[i / 2] = {variants: arr[i + 1].split('~'), state: 0};
//					var maxlength = util.longest_string(inputs[i / 2].variants).length + 2;
					res += '<input type="text" id="var_' + (i / 2) + '" size="' + max_max_length + '" maxlength="' + max_max_length + '" value="' + (state ? state[i / 2] : '') + '" class="wordsenter-input"/>';
				}
				res += arr[len].replace(/\[\[\!--ampdash--\]\]/g, '&#');
				return res;
			},
			attach_events = function (elem) {
				$(elem).find('input').each(function (ind) {

					var i = restrictions.length;
					while (i--) {
						var j = inputs[ind].variants.length, restrict = 1;
						while (j--) {
							if (inputs[ind].variants[j] && inputs[ind].variants[j].match(restrictions[i])) {
								restrict = 0;
								$(this).attr('current_layout', i);
							}
						}
					}

					inputs[ind].el = $(this).on('keyup blur', function (e) {
						var val = this.value, i = restrictions.length;
						while (i--) {
							var j = inputs[ind].variants.length, restrict = 1;
							while (j--) {
								if (inputs[ind].variants[j] && inputs[ind].variants[j].match(restrictions[i])) {
									//val = val.replace(restrictions[i],'');
									restrict = 0;
									current_layout = i;
								}
							}
							if (restrict) {
								val = val.replace(restrictions[i], '');
							}
						}
						this.value = val;
					});
				});
			};
	return {
		/* elem - element, where the data is rendered
		 data - <component> element to render
		 state - current state (if the question is already answered) */
		init_question: function (elem, data, state, r, cb) {
			var font = data.attr('text_size');

			$(elem).html('<div class="wordsenter" ' + (font ? 'style="font-size:' + font + 'px"' : '') + '>' + parse_string(util.get_html(data.find('object')), state) + '</div>');
			attach_events(elem);
			if (util.has_touch()) {
				requirejs(['keyboard'], function () {
					requirejs(['russian'], function () {
						$('.wordsenter input').each(function () {
							$(this).keyboard({autoAccept: true, layout: layouts[$(this).attr('current_layout')]});
						});
					});
				});
			}
			$('input').on('change keyup mouseup touchend', function () {
//				$('#check_btn').addClass('inactive');
				$('input').removeClass('incorrect correct');
				for (var i = 0, l = inputs.length; i < l; i++) {
//					inputs[i].el.removeClass('incorrect correct');
					if (inputs[i].el.val().trim() == '')
						return;
				}
//				$('#check_btn').removeClass('inactive');

				
			});
			if (typeof cb == 'function') {
				cb();
			}
		},
		/* cd - callback, function(result,score,state),
		 result - failed, passed, partial (-1, 1, 0), score - 0-100, state - state to save and resore when returned to this question */
		process_answer: function (cb, forceNotAllChoosedCheck, forceNotDoneCheck) {
			console.log('forceNotAllChoosedCheck ' + forceNotAllChoosedCheck);
			if (typeof forceNotAllChoosedCheck == 'undefined') {
				forceNotAllChoosedCheck = true;
			}
			if (typeof forceNotDoneCheck == 'undefined') {
				forceNotDoneCheck = false;
			}
			var notDone = true, rightAnser = true, notAllChoosed = false;
			for (var i = 0, l = inputs.length; i < l; i++) {
				if (inputs[i].el.val().trim() != '')
					notDone = false;
				else {
					if (!forceNotAllChoosedCheck) {
						notAllChoosed = true;
					}
				}
			}
			if (!forceNotDoneCheck && notDone) {
				return false;
			}
			$('.ui-keyboard').each(function () {
				$(this).remove()
			});
			
			for (var i = 0, l = inputs.length; i < l; i++) {
				var val = inputs[i].el.val().trim().replace(/\s+/g, ' ');
				if (inputs[i].variants.indexOf(val) > -1) {
				} else {
					if (val != '') {
						rightAnser = false;
					}
				}
			}
			
			if (rightAnser && notAllChoosed) {
				cb(0, 0, state);
				return true;
			}
			
			var num_result = 0, state = [];
			for (var i = 0, l = inputs.length; i < l; i++) {
				inputs[i].el.attr('disabled', true);
				var val = inputs[i].el.val().trim().replace(/\s+/g, ' ');
				state[i] = val;
				if (inputs[i].variants.indexOf(val) > -1) {
					inputs[i].state = 1;
					inputs[i].el.addClass('correct');
					inputs[i].el.removeClass('incorrect');
					num_result++;
				} else {
                    inputs[i].el.removeClass('correct incorrect');
                    if (val != '') {
                        inputs[i].el.addClass('incorrect');
                    }
                    inputs[i].state = -1;
					num_result--;
				}
			}
			

			num_result = num_result < 0 ? 0 : num_result * 100 / inputs.length;
			cb(num_result < 100 ? -1 : 1, num_result, state);
			return true;
		},
		show_correct: function () {
			for (var i = 0, l = inputs.length; i < l; i++) {
				inputs[i].el.attr('disabled', true);
				if (inputs[i].state < 1) {
					inputs[i].el.val(inputs[i].variants[0]);
					inputs[i].el.removeClass('incorrect');
					inputs[i].el.removeClass('correct');
				}
			}
		},
		enable_answers: function () {
			for (var i = 0, l = inputs.length; i < l; i++) {
				inputs[i].el.prop('disabled', false);
			}
		},
		disable_answers: function () {
			for (var i = 0, l = inputs.length; i < l; i++) {
				inputs[i].el.prop('disabled', true);
			}
		},
		get_correct_answer_text: function () {
			var curr_task = app_data.cur_task,
					curr_component = app_data.tasks[curr_task].current,
					$nodes = $($(app_data.data.find('task').get(curr_task)).find('component').get(curr_component)).find('object');
			var texts = [];
			$nodes.each(function () {
				texts.push($(this).text().split('~#').join('').split('#').join('').split('~').join(','));
			});
			return texts.join('<br/>');
		},
		get_help_tpl:function(){
			var $tpl = $('.wordsenter-tpls .tpl-helpMessage').clone(true).children();
			
			if($(".wordsenter-input").length > 1){
				$tpl = $tpl.filter('.few-choose').remove();
			}else{
				$tpl = $tpl.filter('.one-choose').remove();
			}
			return $tpl;
			
		}		
	};
});